import json

import pytest

from uniswap_universal_router_decoder._abi_builder import build_abi_type_list  # noqa
from uniswap_universal_router_decoder._enums import (  # noqa
    MiscFunctions,
    RouterFunction,
    V4Actions,
)


expected_fct_abi_08 = json.loads('{"inputs":[{"name":"recipient","type":"address"},{"name":"amountIn","type":"uint256"},{"name":"amountOutMin","type":"uint256"},{"name":"path","type":"address[]"},{"name":"payerIsSender","type":"bool"}],"name":"V2_SWAP_EXACT_IN","type":"function"}')  # noqa: E501
expected_fct_signature_08 = "V2_SWAP_EXACT_IN(address,uint256,uint256,address[],bool)"
expected_fct_abi_10 = json.loads('{"inputs":[{"name":"struct","type":"tuple","components":[{"name":"details","type":"tuple","components":[{"name":"token","type":"address"},{"name":"amount","type":"uint160"},{"name":"expiration","type":"uint48"},{"name":"nonce","type":"uint48"}]},{"name":"spender","type":"address"},{"name":"sigDeadline","type":"uint256"}]},{"name":"data","type":"bytes"}],"name":"PERMIT2_PERMIT","type":"function"}')  # noqa: E501
expected_fct_signature_10 = "PERMIT2_PERMIT(((address,uint160,uint48,uint48),address,uint256),bytes)"
expected_execute_abi = json.loads('{"inputs":[{"name":"commands","type":"bytes"},{"name":"inputs","type":"bytes[]"}],"name":"execute","type":"function"}')  # noqa: E501
expected_execute_signature = "execute(bytes,bytes[])"
expected_fct_abi_19 = json.loads('{"inputs": [{"components": [{"name": "currency0", "type": "address"}, {"name": "currency1", "type": "address"}, {"name": "fee", "type": "uint24"}, {"name": "tickSpacing", "type": "int24"}, {"name": "hooks", "type": "address"}], "name": "PoolKey", "type": "tuple"}, {"name": "sqrtPriceX96", "type": "uint256"}], "name": "V4_INITIALIZE_POOL", "type": "function"}')  # noqa E501
expected_fct_signature_19 = "V4_INITIALIZE_POOL((address,address,uint24,int24,address),uint256)"
expected_v4_swap_exact_in_abi = json.loads('{"inputs": [{"name": "params", "type": "ExactInputParams"}], "name": "SWAP_EXACT_IN", "type": "function"}')  # noqa E501
expected_v4_swap_exact_in_signature = "SWAP_EXACT_IN(ExactInputParams)"


@pytest.mark.parametrize(
    "command_id, expected_fct_abi, expected_selector, expected_signature",
    (
        (RouterFunction(8), expected_fct_abi_08, bytes.fromhex("3bd2d879"), expected_fct_signature_08),
        (RouterFunction(10), expected_fct_abi_10, b'9#\xf7\x04', expected_fct_signature_10),
        (MiscFunctions.EXECUTE, expected_execute_abi, bytes.fromhex("24856bc3"), expected_execute_signature),
        (RouterFunction.V4_INITIALIZE_POOL, expected_fct_abi_19, b'\xdb\xd4\xe4\xbd', expected_fct_signature_19),
        (V4Actions.SWAP_EXACT_IN, expected_v4_swap_exact_in_abi, b'h>~1', expected_v4_swap_exact_in_signature),  # noqa E501
    )
)
def test_build_abi_map(command_id, expected_fct_abi, expected_selector, expected_signature, codec):
    assert codec._abi_map[command_id].get_abi() == expected_fct_abi
    assert codec._abi_map[command_id].get_selector() == expected_selector
    assert codec._abi_map[command_id].get_signature() == expected_signature


abi_dict_1 = \
    {
        'inputs': [
            {
                'name': 'recipient',
                'type': 'address'
            },
            {
                'name': 'amountMin',
                'type': 'uint256'
            }
        ],
        'name': 'WRAP_ETH',
        'type': 'function'
    }

abi_dict_2 = \
    {
        'inputs': [
            {
                'name': 'struct',
                'type': 'tuple',
                'components':
                    [
                        {
                            'name': 'details',
                            'type': 'tuple',
                            'components': [
                                {
                                    'name': 'token',
                                    'type': 'address'
                                },
                                {
                                    'name': 'amount',
                                    'type': 'uint160'
                                },
                                {
                                    'name': 'expiration',
                                    'type': 'uint48'
                                },
                                {
                                    'name': 'nonce',
                                    'type': 'uint48'
                                }
                            ]
                        },
                        {
                            'name': 'spender',
                            'type': 'address'
                        },
                        {
                            'name': 'sigDeadline',
                            'type': 'uint256'
                        }
                    ]
            },
            {
                'name': 'data',
                'type': 'bytes'
            }
        ],
        'name': 'PERMIT2_PERMIT',
        'type': 'function'
    }


@pytest.mark.parametrize(
    "abi_dict, expected_abi_types",
    (
        (abi_dict_1, ["address", "uint256"]),
        (abi_dict_2, ["((address,uint160,uint48,uint48),address,uint256)", "bytes"]),
    )
)
def test_get_abi_types(abi_dict, expected_abi_types):
    assert build_abi_type_list(abi_dict) == expected_abi_types



"""
Encoding part of the Uniswap Universal Router Codec

* Author: Elnaril (https://www.fiverr.com/elnaril, https://github.com/Elnaril).
* License: MIT.
* Doc: https://github.com/Elnaril/uniswap-universal-router-decoder
"""
from __future__ import annotations

from abc import ABC
from typing import (
    Any,
    cast,
    Dict,
    List,
    Optional,
    Sequence,
    TypedDict,
    TypeVar,
    Union,
)

from eth_account.account import SignedMessage
from eth_utils import keccak
from web3 import Web3
from web3.types import (
    BlockIdentifier,
    ChecksumAddress,
    HexStr,
    Nonce,
    TxParams,
    Wei,
)

from uniswap_universal_router_decoder._abi_builder import ABIMap
from uniswap_universal_router_decoder._constants import (
    _router_abi,
    _ur_address,
)
from uniswap_universal_router_decoder._enums import (
    _RouterConstant,
    FunctionRecipient,
    MiscFunctions,
    RouterFunction,
    TransactionSpeed,
    V4Actions,
)
from uniswap_universal_router_decoder.utils import (
    compute_gas_fees,
    compute_sqrt_price_x96,
)


NO_REVERT_FLAG = _RouterConstant.FLAG_ALLOW_REVERT.value


class PoolKey(TypedDict):
    """
    Use v4_pool_key() to make sure currency_0 < currency_1
    """
    currency_0: ChecksumAddress
    currency_1: ChecksumAddress
    fee: int
    tick_spacing: int
    hooks: ChecksumAddress


class PathKey(TypedDict):
    intermediate_currency: ChecksumAddress
    fee: int
    tick_spacing: int
    hooks: ChecksumAddress
    hook_data: bytes


class _Encoder:
    def __init__(self, w3: Web3, abi_map: ABIMap) -> None:
        self._w3 = w3
        self._router_contract = self._w3.eth.contract(abi=_router_abi)
        self._abi_map = abi_map

    @staticmethod
    def v3_path(v3_fn_name: str, path_seq: Sequence[Union[int, ChecksumAddress]]) -> bytes:
        """
        Encode a V3 path
        :param v3_fn_name: 'V3_SWAP_EXACT_IN' or 'V3_SWAP_EXACT_OUT'
        :param path_seq: a sequence of token addresses with the pool fee in between, ex: [tk_in_addr, fee, tk_out_addr]
        :return: the encoded V3 path
        """
        if len(path_seq) < 3:
            raise ValueError("Invalid list to encode a V3 path. Must have at least 3 parameters")
        path_list = list(path_seq)
        if v3_fn_name == "V3_SWAP_EXACT_OUT":
            path_list.reverse()
        elif v3_fn_name != "V3_SWAP_EXACT_IN":
            raise ValueError("v3_fn_name must be in ('V3_SWAP_EXACT_IN', 'V3_SWAP_EXACT_OUT')")
        path = "0x"
        for i, item in enumerate(path_list):
            if i % 2 == 0:
                _item = Web3.to_checksum_address(cast(ChecksumAddress, item))[2:]
            else:
                _item = f"{item:06X}"
            path += _item
        return Web3.to_bytes(hexstr=HexStr(path))

    @staticmethod
    def v4_pool_key(
            currency_0: Union[str, HexStr, ChecksumAddress],
            currency_1: Union[str, HexStr, ChecksumAddress],
            fee: int,
            tick_spacing: int,
            hooks: Union[str, HexStr, ChecksumAddress] = "0x0000000000000000000000000000000000000000") -> PoolKey:
        """
        Make sure currency_0 < currency_1 and returns the v4 pool key

        :param currency_0: the address of one token, or "0x0000000000000000000000000000000000000000" for ETH
        :param currency_1: the address of the other token, or "0x0000000000000000000000000000000000000000" for ETH
        :param fee: pool fee in percentage * 10000 (ex: 3000 for 0.3%)
        :param tick_spacing: granularity of the pool. Lower values are more precise but more expensive to trade
        :param hooks: hook address, default is no hooks, ie "0x0000000000000000000000000000000000000000"
        :return: the v4 pool key
        """
        if int(currency_0, 16) > int(currency_1, 16):
            currency_0, currency_1 = currency_1, currency_0
        return PoolKey(
            currency_0=Web3.to_checksum_address(currency_0),
            currency_1=Web3.to_checksum_address(currency_1),
            fee=int(fee),
            tick_spacing=int(tick_spacing),
            hooks=Web3.to_checksum_address(hooks),
        )

    def v4_pool_id(self, pool_key: PoolKey) -> bytes:
        """
        Encode the pool id

        :param pool_key: the PoolKey (see v4_pool_key() to get it)
        :return: the pool id
        """
        args = (tuple(pool_key.values()), )
        abi = self._abi_map[MiscFunctions.V4_POOL_ID]
        return keccak(abi.encode(args))

    @staticmethod
    def v4_path_key(
            intermediate_currency: ChecksumAddress,
            fee: int,
            tick_spacing: int,
            hooks: Union[str, HexStr, ChecksumAddress] = "0x0000000000000000000000000000000000000000",
            hook_data: bytes = b"") -> PathKey:
        """
        Build a PathKey which is used by multi-hop swap encoding

        :param intermediate_currency: the address of one token of the target pool
        :param fee: pool fee in percentage * 10000 (ex: 3000 for 0.3%)
        :param tick_spacing: granularity of the pool. Lower values are more precise but more expensive to trade
        :param hooks: hook address, default is no hooks, ie "0x0000000000000000000000000000000000000000"
        :param hook_data: encoded hook data
        :return: the corresponding PathKey
        """
        return PathKey(
            intermediate_currency=Web3.to_checksum_address(intermediate_currency),
            fee=int(fee),
            tick_spacing=int(tick_spacing),
            hooks=Web3.to_checksum_address(hooks),
            hook_data=hook_data,
        )

    def chain(self) -> _ChainedFunctionBuilder:
        """
        :return: Initialize the chain of encoded functions
        """
        return _ChainedFunctionBuilder(self._w3, self._abi_map)


TV4ChainedCommonFunctionBuilder = TypeVar("TV4ChainedCommonFunctionBuilder", bound="_V4ChainedCommonFunctionBuilder")


class _V4ChainedCommonFunctionBuilder(ABC):
    def __init__(self, builder: _ChainedFunctionBuilder, w3: Web3, abi_map: ABIMap):
        self.builder = builder
        self._w3 = w3
        self._abi_map = abi_map
        self.actions: bytearray = bytearray()
        self.arguments: List[bytes] = []

    def _add_action(self, action: V4Actions, args: Sequence[Any]) -> None:
        abi = self._abi_map[action]
        self.actions.append(action.value)
        self.arguments.append(abi.encode(args))

    def settle(
            self: TV4ChainedCommonFunctionBuilder,
            currency: ChecksumAddress,
            amount: int,
            payer_is_user: bool) -> TV4ChainedCommonFunctionBuilder:
        """
        Pay the contract for a given amount of tokens (currency). Used for swaps and position management.

        :param currency: The token addr or "0x0000000000000000000000000000000000000000" for the native currency (ie ETH)
        :param amount: The amount to send to the contract
        :param payer_is_user: Whether this amount comes from the transaction sender or not.
        :return: The chain link corresponding to this function call.
        """
        args = (currency, amount, payer_is_user)
        self._add_action(V4Actions.SETTLE, args)
        return self

    def take(
            self: TV4ChainedCommonFunctionBuilder,
            currency: ChecksumAddress,
            recipient: ChecksumAddress,
            amount: int) -> TV4ChainedCommonFunctionBuilder:
        """
        Get the given amount of tokens (currency) from the contract. Used for swaps and position management.

        :param currency: The token addr or "0x0000000000000000000000000000000000000000" for the native currency (ie ETH)
        :param recipient: Address of who gets the tokens
        :param amount: The token amount to get
        :return: The chain link corresponding to this function call.
        """
        args = (currency, recipient, amount)
        self._add_action(V4Actions.TAKE, args)
        return self


class _V4ChainedPositionFunctionBuilder(_V4ChainedCommonFunctionBuilder):

    def mint_position(
            self,
            pool_key: PoolKey,
            tick_lower: int,
            tick_upper: int,
            liquidity: int,
            amount_0_max: int,
            amount_1_max: int,
            recipient: ChecksumAddress,
            hook_data: bytes) -> _V4ChainedPositionFunctionBuilder:
        """
        Position - Mint a V4 position, ie add some liquidity to a given pool and get a position as ERC-721 token.

        :param pool_key: The PoolKey to identify the pool where the liquidity will be added to.
        :param tick_lower: The lower tick boundary of the position
        :param tick_upper: The upper tick boundary of the position
        :param liquidity: The amount of liquidity units to mint
        :param amount_0_max: The maximum amount of currency_0 the sender is willing to pay
        :param amount_1_max: The maximum amount of currency_1 the sender is willing to pay
        :param recipient: The address that will receive the liquidity position (ERC-721)
        :param hook_data: The encoded hook data to be forwarded to the hook functions
        :return: The chain link corresponding to this function call.
        """
        args = (
            tuple(pool_key.values()),
            tick_lower,
            tick_upper,
            liquidity,
            amount_0_max,
            amount_1_max,
            recipient,
            hook_data
        )
        self._add_action(V4Actions.MINT_POSITION, args)
        return self

    def settle_pair(
            self,
            currency_0: ChecksumAddress,
            currency_1: ChecksumAddress) -> _V4ChainedPositionFunctionBuilder:
        """
        Position - Indicates that tokens are to be paid by the caller to create the position.

        :param currency_0: The address of 1 token
        :param currency_1: The address of the other token
        :return: The chain link corresponding to this function call.
        """
        args = (currency_0, currency_1)
        self._add_action(V4Actions.SETTLE_PAIR, args)
        return self

    def close_currency(self, currency: ChecksumAddress) -> _V4ChainedPositionFunctionBuilder:
        """
        Position - Automatically determines if a currency should be settled or taken.

        :param currency: The token or ETH to be paid or received.
        :return: The chain link corresponding to this function call.
        """
        args = (currency, )
        self._add_action(V4Actions.CLOSE_CURRENCY, args)
        return self

    def sweep(self, currency: ChecksumAddress, to: ChecksumAddress) -> _V4ChainedPositionFunctionBuilder:
        """
        Position - Sweep the contract

        :param currency: The token or ETH address to sweep
        :param to: The sweep recipient address
        :return: The chain link corresponding to this function call.
        """
        args = (currency, to)
        self._add_action(V4Actions.SWEEP, args)
        return self

    # def mint_position_from_deltas(
    #         self,
    #         pool_key: PoolKey,
    #         tick_lower: int,
    #         tick_upper: int,
    #         amount_0_max: int,
    #         amount_1_max: int,
    #         recipient: ChecksumAddress,
    #         hook_data: bytes) -> _V4ChainedPositionFunctionBuilder:
    #     args = (
    #         tuple(pool_key.values()),
    #         tick_lower,
    #         tick_upper,
    #         amount_0_max,
    #         amount_1_max,
    #         recipient,
    #         hook_data
    #     )
    #     self._add_action(V4Actions.MINT_POSITION_FROM_DELTAS, args)
    #     return self

    def wrap_eth(self, amount: Wei) -> _V4ChainedPositionFunctionBuilder:
        """
        Position - Encode the call to the function WRAP which convert ETH to WETH in the position manager contract

        :param amount: The amount of ETH in Wei.
        :return: The chain link corresponding to this function call.
        """
        args = (amount, )
        self._add_action(V4Actions.WRAP, args)
        return self

    def unwrap_weth(self, amount: Wei) -> _V4ChainedPositionFunctionBuilder:
        """
        Position - Encode the call to the function UNWRAP which convert WETH to ETH in the position manager contract

        :param amount: The amount of WETH in Wei.
        :return: The chain link corresponding to this function call.
        """
        args = (amount, )
        self._add_action(V4Actions.UNWRAP, args)
        return self

    def take_pair(
            self,
            currency_0: ChecksumAddress,
            currency_1: ChecksumAddress,
            recipient: ChecksumAddress) -> _V4ChainedPositionFunctionBuilder:
        args = (currency_0, currency_1, recipient)
        self._add_action(V4Actions.TAKE_PAIR, args)
        return self

    def clear_or_take(
            self,
            currency: ChecksumAddress,
            amount_max: int) -> _V4ChainedPositionFunctionBuilder:
        """
        Position - If the token amount to-be-collected is below a threshold, opt to forfeit the dust.
        Otherwise, claim the tokens.

        :param currency: The token or ETH address to be forfeited or claimed.
        :param amount_max: The threshold.
        :return: The chain link corresponding to this function call.
        """
        args = (currency, amount_max)
        self._add_action(V4Actions.CLEAR_OR_TAKE, args)
        return self

    def build_v4_posm_call(self, deadline: int) -> _ChainedFunctionBuilder:
        """
        Position - Build the V4 position manager call

        :param deadline: When this call is not valid anymore
        :return: The chain link corresponding to this function call.
        """
        action_values = (bytes(self.actions), self.arguments)
        abi = self._abi_map[MiscFunctions.UNLOCK_DATA]
        encoded_data = abi.encode(action_values)
        args = (encoded_data, deadline)
        self.builder._add_command(RouterFunction.V4_POSITION_MANAGER_CALL, args, True)
        return self.builder


class _V4ChainedSwapFunctionBuilder(_V4ChainedCommonFunctionBuilder):

    def swap_exact_in_single(
            self,
            pool_key: PoolKey,
            zero_for_one: bool,
            amount_in: Wei,
            amount_out_min: Wei,
            hook_data: bytes = b'') -> _V4ChainedSwapFunctionBuilder:
        """
        Swap - Encode the call to the V4_SWAP function SWAP_EXACT_IN_SINGLE.

        :param pool_key: the target pool key returned by encode.v4_pool_key()
        :param zero_for_one: the swap direction, true for currency_0 to currency_1, false for currency_1 to currency_0
        :param amount_in: the exact amount of the sold currency in Wei
        :param amount_out_min: the minimum accepted bought currency
        :param hook_data: encoded data sent to the pool's hook, if any.
        :return: The chain link corresponding to this function call.
        """
        args = ((tuple(pool_key.values()), zero_for_one, amount_in, amount_out_min, hook_data),)
        self._add_action(V4Actions.SWAP_EXACT_IN_SINGLE, args)
        return self

    def take_all(self, currency: ChecksumAddress, min_amount: Wei) -> _V4ChainedSwapFunctionBuilder:
        """
        Swap - Final action that collects all output tokens after the swap is complete.

        :param currency: The address of the token or ETH ("0x0000000000000000000000000000000000000000") to receive
        :param min_amount: The expected minimum amount to be received.
        :return: The chain link corresponding to this function call.
        """
        args = (currency, min_amount)
        self._add_action(V4Actions.TAKE_ALL, args)
        return self

    def settle_all(self, currency: ChecksumAddress, max_amount: Wei) -> _V4ChainedSwapFunctionBuilder:
        """
        Swap - Final action that ensures all input tokens involved in the swap are properly paid to the contract.

        :param currency: The address of the token or ETH ("0x0000000000000000000000000000000000000000") to pay.
        :param max_amount: The expected maximum amount to pay.
        :return: The chain link corresponding to this function call.
        """
        args = (currency, max_amount)
        self._add_action(V4Actions.SETTLE_ALL, args)
        return self

    def swap_exact_in(
            self,
            currency_in: ChecksumAddress,
            path_keys: Sequence[PathKey],
            amount_in: int,
            amount_out_min: int) -> _V4ChainedSwapFunctionBuilder:
        """
        Swap - Encode Multi-hop V4 SWAP_EXACT_IN swaps.

        :param currency_in: The address of the token (or ETH) to send to the first pool.
        :param path_keys: The sequence of path keys to identify the pools
        :param amount_in: The amount to send to the contract
        :param amount_out_min: The expected minimum amount to be received
        :return: The chain link corresponding to this function call.
        """
        args = ((currency_in, [tuple(path_key.values()) for path_key in path_keys], amount_in, amount_out_min), )
        self._add_action(V4Actions.SWAP_EXACT_IN, args)
        return self

    def swap_exact_out_single(
            self,
            pool_key: PoolKey,
            zero_for_one: bool,
            amount_out: Wei,
            amount_in_max: Wei,
            hook_data: bytes = b'') -> _V4ChainedSwapFunctionBuilder:
        """
        Swap - Encode the call to the V4_SWAP function SWAP_EXACT_IN_SINGLE.

        :param pool_key: the target pool key returned by encode.v4_pool_key()
        :param zero_for_one: the swap direction, true for currency_0 to currency_1, false for currency_1 to currency_0
        :param amount_out: the exact amount of the bought currency in Wei
        :param amount_in_max: the maximum accepted sold currency
        :param hook_data: encoded data sent to the pool's hook, if any.
        :return: The chain link corresponding to this function call.
        """
        args = ((tuple(pool_key.values()), zero_for_one, amount_out, amount_in_max, hook_data),)
        self._add_action(V4Actions.SWAP_EXACT_OUT_SINGLE, args)
        return self

    def swap_exact_out(
            self,
            currency_out: ChecksumAddress,
            path_keys: Sequence[PathKey],
            amount_out: int,
            amount_in_max: int) -> _V4ChainedSwapFunctionBuilder:
        """
        Swap - Encode Multi-hop V4 SWAP_EXACT_OUT swaps.

        :param currency_out: The address of the token (or ETH) to be received
        :param path_keys: The sequence of path keys to identify the pools
        :param amount_out: The amount to be received
        :param amount_in_max: The maximum amount that the transaction sender is willing to pay.
        :return: The chain link corresponding to this function call.
        """
        args = ((currency_out, [tuple(path_key.values()) for path_key in path_keys], amount_out, amount_in_max), )
        self._add_action(V4Actions.SWAP_EXACT_OUT, args)
        return self

    def take_portion(
            self,
            currency: ChecksumAddress,
            recipient: ChecksumAddress,
            bips: int) -> _V4ChainedSwapFunctionBuilder:
        """
        Swap - Send a portion of token to a given recipient.

        :param currency: The address of the token (or ETH) to use for this action
        :param recipient: The address which will receive the tokens
        :param bips: The recipient will receive balance * bips / 10_000
        :return: The chain link corresponding to this function call.
        """
        args = (currency, recipient, bips)
        self._add_action(V4Actions.TAKE_PORTION, args)
        return self

    def build_v4_swap(self) -> _ChainedFunctionBuilder:
        """
        Build the V4 swap call

        :return: The chain link corresponding to this function call.
        """
        args = (bytes(self.actions), self.arguments)
        self.builder._add_command(RouterFunction.V4_SWAP, args)
        return self.builder


class _ChainedFunctionBuilder:
    def __init__(self, w3: Web3, abi_map: ABIMap):
        self._w3 = w3
        self._router_contract = self._w3.eth.contract(abi=_router_abi)
        self._abi_map = abi_map
        self.commands: bytearray = bytearray()
        self.arguments: List[bytes] = []

    def _add_command(self, command: RouterFunction, args: Sequence[Any], add_selector: bool = False) -> None:
        abi = self._abi_map[command]
        self.commands.append(command.value)
        arguments = abi.get_selector() + abi.encode(args) if add_selector else abi.encode(args)
        self.arguments.append(arguments)

    @staticmethod
    def _get_recipient(
            function_recipient: FunctionRecipient,
            custom_recipient: Optional[ChecksumAddress] = None) -> ChecksumAddress:
        recipient_mapping = {
            FunctionRecipient.SENDER: _RouterConstant.MSG_SENDER.value,
            FunctionRecipient.ROUTER: _RouterConstant.ADDRESS_THIS.value,
            FunctionRecipient.CUSTOM: custom_recipient,
        }
        recipient = recipient_mapping[function_recipient]
        if recipient:
            return Web3.to_checksum_address(recipient)
        else:
            raise ValueError(
                f"Invalid function_recipient: {function_recipient} or custom_recipient: {custom_recipient}: "
                f"custom_recipient address must be provided if FunctionRecipient.CUSTOM is selected."
            )

    @staticmethod
    def _get_command(router_function: RouterFunction, revert_on_fail: bool) -> int:
        return int(router_function.value if revert_on_fail else router_function.value | NO_REVERT_FLAG)

    def wrap_eth(
            self,
            function_recipient: FunctionRecipient,
            amount: Wei,
            custom_recipient: Optional[ChecksumAddress] = None) -> _ChainedFunctionBuilder:
        """
        Encode the call to the function WRAP_ETH which convert ETH to WETH through the UR

        :param function_recipient: A FunctionRecipient which defines the recipient of this function output.
        :param amount: The amount of sent ETH in WEI.
        :param custom_recipient: If function_recipient is CUSTOM, must be the actual recipient, otherwise None.

        :return: The chain link corresponding to this function call.
        """
        recipient = self._get_recipient(function_recipient, custom_recipient)
        args = (recipient, amount)
        self._add_command(RouterFunction.WRAP_ETH, args)
        return self

    def unwrap_weth(
            self,
            function_recipient: FunctionRecipient,
            amount: Wei,
            custom_recipient: Optional[ChecksumAddress] = None) -> _ChainedFunctionBuilder:
        """
        Encode the call to the function UNWRAP_WETH which convert WETH to ETH through the UR

        :param function_recipient: A FunctionRecipient which defines the recipient of this function output.
        :param amount: The amount of sent WETH in WEI.
        :param custom_recipient: If function_recipient is CUSTOM, must be the actual recipient, otherwise None.

        :return: The chain link corresponding to this function call.
        """
        recipient = self._get_recipient(function_recipient, custom_recipient)
        args = (recipient, amount)
        self._add_command(RouterFunction.UNWRAP_WETH, args)
        return self

    def v2_swap_exact_in(
            self,
            function_recipient: FunctionRecipient,
            amount_in: Wei,
            amount_out_min: Wei,
            path: Sequence[ChecksumAddress],
            custom_recipient: Optional[ChecksumAddress] = None,
            payer_is_sender: bool = True) -> _ChainedFunctionBuilder:
        """
        Encode the call to the function V2_SWAP_EXACT_IN, which swaps tokens on Uniswap V2.
        Correct allowances must have been set before sending such transaction.

        :param function_recipient: A FunctionRecipient which defines the recipient of this function output.
        :param amount_in: The exact amount of the sold (token_in) token in Wei
        :param amount_out_min: The minimum accepted bought token (token_out)
        :param path: The V2 path: a list of 2 or 3 tokens where the first is token_in and the last is token_out
        :param custom_recipient: If function_recipient is CUSTOM, must be the actual recipient, otherwise None.
        :param payer_is_sender: True if the in tokens come from the sender, False if they already are in the router

        :return: The chain link corresponding to this function call.
        """
        recipient = self._get_recipient(function_recipient, custom_recipient)
        args = (recipient, amount_in, amount_out_min, path, payer_is_sender)
        self._add_command(RouterFunction.V2_SWAP_EXACT_IN, args)
        return self

    def v2_swap_exact_in_from_balance(
            self,
            function_recipient: FunctionRecipient,
            amount_out_min: Wei,
            path: Sequence[ChecksumAddress],
            custom_recipient: Optional[ChecksumAddress] = None) -> _ChainedFunctionBuilder:
        """
        Encode the call to the function V2_SWAP_EXACT_IN, using the router balance as amount_in,
        which swaps tokens on Uniswap V2.
        Typically used when the amount_in is unknown because it comes from a V*_SWAP_EXACT_IN output.
        Correct allowances must have been set before sending such transaction.

        :param function_recipient: A FunctionRecipient which defines the recipient of this function output.
        :param amount_out_min: The minimum accepted bought token (token_out)
        :param path: The V2 path: a list of 2 or 3 tokens where the first is token_in and the last is token_out
        :param custom_recipient: If function_recipient is CUSTOM, must be the actual recipient, otherwise None.

        :return: The chain link corresponding to this function call.
        """
        return self.v2_swap_exact_in(
            function_recipient,
            _RouterConstant.ROUTER_BALANCE.value,
            amount_out_min,
            path,
            custom_recipient,
            False,
        )

    def v2_swap_exact_out(
            self,
            function_recipient: FunctionRecipient,
            amount_out: Wei,
            amount_in_max: Wei,
            path: Sequence[ChecksumAddress],
            custom_recipient: Optional[ChecksumAddress] = None,
            payer_is_sender: bool = True) -> _ChainedFunctionBuilder:
        """
        Encode the call to the function V2_SWAP_EXACT_OUT, which swaps tokens on Uniswap V2.
        Correct allowances must have been set before sending such transaction.

        :param function_recipient: A FunctionRecipient which defines the recipient of this function output.
        :param amount_out: The exact amount of the bought (token_out) token in Wei
        :param amount_in_max: The maximum accepted sold token (token_in)
        :param path: The V2 path: a list of 2 or 3 tokens where the first is token_in and the last is token_out
        :param custom_recipient: If function_recipient is CUSTOM, must be the actual recipient, otherwise None.
        :param payer_is_sender: True if the in tokens come from the sender, False if they already are in the router

        :return: The chain link corresponding to this function call.
        """
        recipient = self._get_recipient(function_recipient, custom_recipient)
        args = (recipient, amount_out, amount_in_max, path, payer_is_sender)
        self._add_command(RouterFunction.V2_SWAP_EXACT_OUT, args)
        return self

    def v3_swap_exact_in(
            self,
            function_recipient: FunctionRecipient,
            amount_in: Wei,
            amount_out_min: Wei,
            path: Sequence[Union[int, ChecksumAddress]],
            custom_recipient: Optional[ChecksumAddress] = None,
            payer_is_sender: bool = True) -> _ChainedFunctionBuilder:
        """
        Encode the call to the function V3_SWAP_EXACT_IN, which swaps tokens on Uniswap V3.
        Correct allowances must have been set before sending such transaction.

        :param function_recipient: A FunctionRecipient which defines the recipient of this function output.
        :param amount_in: The exact amount of the sold (token_in) token in Wei
        :param amount_out_min: The minimum accepted bought token (token_out) in Wei
        :param path: The V3 path: a list of tokens where the first is the token_in, the last one is the token_out, and
        with the pool fee between each token in percentage * 10000 (ex: 3000 for 0.3%)
        :param custom_recipient: If function_recipient is CUSTOM, must be the actual recipient, otherwise None.
        :param payer_is_sender: True if the in tokens come from the sender, False if they already are in the router

        :return: The chain link corresponding to this function call.
        """
        recipient = self._get_recipient(function_recipient, custom_recipient)
        encoded_v3_path = _Encoder.v3_path(RouterFunction.V3_SWAP_EXACT_IN.name, path)
        args = (recipient, amount_in, amount_out_min, encoded_v3_path, payer_is_sender)
        self._add_command(RouterFunction.V3_SWAP_EXACT_IN, args)
        return self

    def v3_swap_exact_in_from_balance(
            self,
            function_recipient: FunctionRecipient,
            amount_out_min: Wei,
            path: Sequence[Union[int, ChecksumAddress]],
            custom_recipient: Optional[ChecksumAddress] = None) -> _ChainedFunctionBuilder:
        """
        Encode the call to the function V3_SWAP_EXACT_IN, using the router balance as amount_in,
        which swaps tokens on Uniswap V3.
        Typically used when the amount_in is unknown because it comes from a V*_SWAP_EXACT_IN output.
        Correct allowances must have been set before sending such transaction.

        :param function_recipient: A FunctionRecipient which defines the recipient of this function output.
        :param amount_out_min: The minimum accepted bought token (token_out) in Wei
        :param path: The V3 path: a list of tokens where the first is the token_in, the last one is the token_out, and
        with the pool fee between each token in percentage * 10000 (ex: 3000 for 0.3%)
        :param custom_recipient: If function_recipient is CUSTOM, must be the actual recipient, otherwise None.

        :return: The chain link corresponding to this function call.
        """
        return self.v3_swap_exact_in(
            function_recipient,
            _RouterConstant.ROUTER_BALANCE.value,
            amount_out_min,
            path,
            custom_recipient,
            False,
        )

    def v3_swap_exact_out(
            self,
            function_recipient: FunctionRecipient,
            amount_out: Wei,
            amount_in_max: Wei,
            path: Sequence[Union[int, ChecksumAddress]],
            custom_recipient: Optional[ChecksumAddress] = None,
            payer_is_sender: bool = True) -> _ChainedFunctionBuilder:
        """
        Encode the call to the function V3_SWAP_EXACT_OUT, which swaps tokens on Uniswap V3.
        Correct allowances must have been set before sending such transaction.

        :param function_recipient: A FunctionRecipient which defines the recipient of this function output.
        :param amount_out: The exact amount of the bought (token_out) token in Wei
        :param amount_in_max: The maximum accepted sold token (token_in) in Wei
        :param path: The V3 path: a list of tokens where the first is the token_in, the last one is the token_out, and
        with the pool fee between each token in percentage * 10000 (ex: 3000 for 0.3%)
        :param custom_recipient: If function_recipient is CUSTOM, must be the actual recipient, otherwise None.
        :param payer_is_sender: True if the in tokens come from the sender, False if they already are in the router

        :return: The chain link corresponding to this function call.
        """
        recipient = self._get_recipient(function_recipient, custom_recipient)
        encoded_v3_path = _Encoder.v3_path(RouterFunction.V3_SWAP_EXACT_OUT.name, path)
        args = (recipient, amount_out, amount_in_max, encoded_v3_path, payer_is_sender)
        self._add_command(RouterFunction.V3_SWAP_EXACT_OUT, args)
        return self

    def permit2_permit(
            self,
            permit_single: Dict[str, Any],
            signed_permit_single: SignedMessage) -> _ChainedFunctionBuilder:
        """
        Encode the call to the function PERMIT2_PERMIT, which gives token allowances to the Permit2 contract.
        In addition, the Permit2 must be approved using the token contracts as usual.

        :param permit_single: The 1st element returned by create_permit2_signable_message()
        :param signed_permit_single: The 2nd element returned by create_permit2_signable_message(), once signed.

        :return: The chain link corresponding to this function call.
        """
        struct = (
            tuple(permit_single["details"].values()),
            permit_single["spender"],
            permit_single["sigDeadline"],
        )
        args = (struct, signed_permit_single.signature)
        self._add_command(RouterFunction.PERMIT2_PERMIT, args)
        return self

    def sweep(
            self,
            function_recipient: FunctionRecipient,
            token_address: ChecksumAddress,
            amount_min: Wei,
            custom_recipient: Optional[ChecksumAddress] = None) -> _ChainedFunctionBuilder:
        """
        Encode the call to the function SWEEP which sweeps all of the router's ERC20 or ETH to an address

        :param function_recipient: A FunctionRecipient which defines the recipient of this function output.
        :param token_address: The address of the token to sweep or "0x0000000000000000000000000000000000000000" for ETH.
        :param amount_min: The minimum desired amount
        :param custom_recipient: If function_recipient is CUSTOM, must be the actual recipient, otherwise None.

        :return: The chain link corresponding to this function call.
        """
        recipient = self._get_recipient(function_recipient, custom_recipient)
        args = (token_address, recipient, amount_min)
        self._add_command(RouterFunction.SWEEP, args)
        return self

    def pay_portion(
            self,
            function_recipient: FunctionRecipient,
            token_address: ChecksumAddress,
            bips: int,
            custom_recipient: Optional[ChecksumAddress] = None) -> _ChainedFunctionBuilder:
        """
        Encode the call to the function PAY_PORTION which transfers a part of the router's ERC20 or ETH to an address.
        Transferred amount = balance * bips / 10_000

        :param function_recipient: A FunctionRecipient which defines the recipient of this function output.
        :param token_address: The address of token to pay or "0x0000000000000000000000000000000000000000" for ETH.
        :param bips: integer between 0 and 10_000
        :param custom_recipient: If function_recipient is CUSTOM, must be the actual recipient, otherwise None.

        :return: The chain link corresponding to this function call.
        """
        if (
            bips < 0
            or bips > 10_000
            or not isinstance(bips, int)
        ):
            raise ValueError(f"Invalid argument: bips must be an int between 0 and 10_000. Received {bips}")

        recipient = self._get_recipient(function_recipient, custom_recipient)
        args = (token_address, recipient, bips)
        self._add_command(RouterFunction.PAY_PORTION, args)
        return self

    def transfer(
            self,
            function_recipient: FunctionRecipient,
            token_address: ChecksumAddress,
            value: Wei,
            custom_recipient: Optional[ChecksumAddress] = None) -> _ChainedFunctionBuilder:
        """
        Encode the call to the function TRANSFER which transfers an amount of ERC20 or ETH from the router's balance
        to an address.

        :param function_recipient: A FunctionRecipient which defines the recipient of this function output.
        :param token_address: The address of token to pay or "0x0000000000000000000000000000000000000000" for ETH.
        :param value: The amount to transfer (in Wei)
        :param custom_recipient: If function_recipient is CUSTOM, must be the actual recipient, otherwise None.

        :return: The chain link corresponding to this function call.
        """

        recipient = self._get_recipient(function_recipient, custom_recipient)
        args = (token_address, recipient, value)
        self._add_command(RouterFunction.TRANSFER, args)
        return self

    def permit2_transfer_from(
            self,
            function_recipient: FunctionRecipient,
            token_address: ChecksumAddress,
            amount: Wei,
            custom_recipient: Optional[ChecksumAddress] = None) -> _ChainedFunctionBuilder:
        """
        Encode the transfer of tokens from the caller address to the given recipient.
        The UR must have been permit2'ed for the token first.

        :param function_recipient: A FunctionRecipient which defines the recipient of this function output.
        :param token_address: The address of the token to be transferred.
        :param amount: The amount to transfer.
        :param custom_recipient: If function_recipient is CUSTOM, must be the actual recipient, otherwise None.
        :return: The chain link corresponding to this function call.
        """
        recipient = self._get_recipient(function_recipient, custom_recipient)
        args = (token_address, recipient, amount)
        self._add_command(RouterFunction.PERMIT2_TRANSFER_FROM, args)
        return self

    def v4_swap(self) -> _V4ChainedSwapFunctionBuilder:
        """
        V4 - Start building a call to the V4 swap functions

        :return: The chain link corresponding to this function call.
        """
        return _V4ChainedSwapFunctionBuilder(self, self._w3, self._abi_map)

    def v4_initialize_pool(self, pool_key: PoolKey, amount_0: Wei, amount_1: Wei) -> _ChainedFunctionBuilder:
        """
        V4 - Encode the call to initialize (create) a V4 pool.
        The amounts are used to compute the initial sqrtPriceX96. They are NOT sent.
        So, basically only the ratio amount_0 / amount_1 is relevant.
        For ex, to create a pool with 2 USD stable coins, you just need to set amount_0 = amount_1 = 1

        :param pool_key: The pool key that identify the pool to create
        :param amount_0: "virtual" amount of PoolKey.currency_0
        :param amount_1: "virtual" amount of PoolKey.currency_1
        :return: The chain link corresponding to this function call.
        """
        sqrt_price_x96 = compute_sqrt_price_x96(amount_0, amount_1)
        args = (tuple(pool_key.values()), sqrt_price_x96)
        self._add_command(RouterFunction.V4_INITIALIZE_POOL, args)
        return self

    def v4_posm_call(self) -> _V4ChainedPositionFunctionBuilder:
        """
        V4 - Start building a call to the V4 positon manager functions
        :return: The chain link corresponding to this function call.
        """
        return _V4ChainedPositionFunctionBuilder(self, self._w3, self._abi_map)

    def build(self, deadline: Optional[int] = None) -> HexStr:
        """
        Build the encoded input for all the chained commands, ready to be sent to the UR

        :param deadline: The optional unix timestamp after which the transaction won't be valid anymore.
        :return: The encoded data to add to the UR transaction dictionary parameters.
        """
        if deadline:
            execute_with_deadline_args = (bytes(self.commands), self.arguments, deadline)
            abi = self._abi_map[MiscFunctions.EXECUTE_WITH_DEADLINE]
            return Web3.to_hex(abi.get_selector() + abi.encode(execute_with_deadline_args))
        else:
            execute_args = (bytes(self.commands), self.arguments)
            abi = self._abi_map[MiscFunctions.EXECUTE]
            return Web3.to_hex(abi.get_selector() + abi.encode(execute_args))

    def build_transaction(
            self,
            sender: ChecksumAddress,
            value: Wei = Wei(0),
            trx_speed: Optional[TransactionSpeed] = TransactionSpeed.FAST,
            *,
            priority_fee: Optional[Wei] = None,
            max_fee_per_gas: Optional[Wei] = None,
            max_fee_per_gas_limit: Wei = Wei(100 * 10 ** 9),
            gas_limit: Optional[int] = None,
            chain_id: Optional[int] = None,
            nonce: Optional[Union[int, Nonce]] = None,
            ur_address: ChecksumAddress = _ur_address,
            deadline: Optional[int] = None,
            block_identifier: BlockIdentifier = "latest") -> TxParams:
        """
        Build the encoded data and the transaction dictionary, ready to be signed.

        By default, compute the gas fees, chain_id, nonce, deadline, ... but custom values can be used instead.
        Gas fees are computed with a given TransactionSpeed (default is FAST).
        All speeds will compute gas fees in order to try to place the transaction in the next block
        (without certainty, of course).
        So, during strained conditions, the computed gas fees could be very high but can be controlled
        thanks to 'max_fee_per_gas_limit'.

        Either a transaction speed is provided or custom gas fees, otherwise a ValueError is raised.

        The RouterCodec must be built with a Web3 instance or a rpc endpoint address except if custom values are used.

        :param sender: The 'from' field - Mandatory
        :param value: The quantity of ETH sent to the Universal Router - Default is 0
        :param trx_speed: The indicative 'speed' of the transaction - Default is TransactionSpeed.FAST
        :param priority_fee: custom 'maxPriorityFeePerGas' - Default is None
        :param max_fee_per_gas: custom 'maxFeePerGas' - Default is None
        :param max_fee_per_gas_limit: if the computed 'max_fee_per_gas' is greater than 'max_fee_per_gas_limit', raise a ValueError  # noqa
        :param gas_limit: custom 'gas' - Default is None
        :param chain_id: custom 'chainId'
        :param nonce: custom 'nonce'
        :param ur_address: custom Universal Router address
        :param deadline: The optional unix timestamp after which the transaction won't be valid anymore.
        :param block_identifier: specify at what block the computing is done. Mostly for test purposes.
        :return: a transaction (TxParams) ready to be signed
        """
        encoded_data = self.build(deadline)

        if chain_id is None:
            chain_id = self._w3.eth.chain_id

        if nonce is None:
            nonce = self._w3.eth.get_transaction_count(sender, block_identifier)

        if not trx_speed:
            if priority_fee is None or max_fee_per_gas is None:
                raise ValueError("Either trx_speed or both priority_fee and max_fee_per_gas must be set.")
            else:
                _priority_fee = priority_fee
                _max_fee_per_gas = max_fee_per_gas
        else:
            if priority_fee or max_fee_per_gas:
                raise ValueError("priority_fee and max_fee_per_gas can't be set with trx_speed")
            else:
                _priority_fee, _max_fee_per_gas = compute_gas_fees(self._w3, trx_speed, block_identifier)
                if _max_fee_per_gas > max_fee_per_gas_limit:
                    raise ValueError(
                        "Computed max_fee_per_gas is greater than max_fee_per_gas_limit. "
                        "Either provide max_fee_per_gas, increase max_fee_per_gas_limit "
                        "or wait for less strained conditions"
                    )

        tx_params: TxParams = {
            "from": sender,
            "value": value,
            "to": ur_address,
            "chainId": chain_id,
            "nonce": Nonce(nonce),
            "type": HexStr('0x2'),
            "maxPriorityFeePerGas": _priority_fee,
            "maxFeePerGas": _max_fee_per_gas,
            "data": encoded_data,
        }

        if gas_limit is None:
            estimated_gas = self._w3.eth.estimate_gas(tx_params, block_identifier)
            gas_limit = int(estimated_gas * 1.15)

        tx_params["gas"] = Wei(gas_limit)

        return tx_params



import pytest

from uniswap_universal_router_decoder import TransactionSpeed
from uniswap_universal_router_decoder.utils import compute_gas_fees


@pytest.mark.parametrize(
    "trx_speed, block_identifier, expected_priority_fee, expected_max_fee_per_gas",
    (
        (TransactionSpeed.SLOW, 19371376, 51749500, 267500297642),
        (TransactionSpeed.AVERAGE, 19371376, 1336613173, 268785161315),
        (TransactionSpeed.FAST, 19371376, 2500000000, 269948548142),
        (TransactionSpeed.FASTER, 19371376, 3000000000, 270448548142),
        (TransactionSpeed.FAST, 19874484, 2500000000, 11746497877),
        (TransactionSpeed.FASTER, 19874484, 3000000000, 12246497877),
        (TransactionSpeed.FAST, 12965000, 1, 1500000001),  # EIP-1559 went live at block 12965000
    )
)
def test_compute_gas_fees(trx_speed, block_identifier, expected_priority_fee, expected_max_fee_per_gas, w3):
    priority_fee, max_fee_per_gas = compute_gas_fees(w3, trx_speed, block_identifier)
    assert priority_fee == expected_priority_fee
    assert max_fee_per_gas == expected_max_fee_per_gas



import pytest
from web3 import Web3


def test_get_default_deadline(codec):
    assert 79 < codec.get_default_deadline() - codec.get_default_deadline(100) < 81


def test_get_default_expiration(codec):
    assert -1 < codec.get_default_expiration() - codec.get_default_expiration(30*24*3600) < 1


def test_get_max_expiration(codec):
    assert codec.get_max_expiration() == 2**48 - 1


@pytest.mark.parametrize(
    "wallet, token, block_identifier, expected_result",
    (
        (Web3.to_checksum_address('0x1944922D86209F7e5F3c88F2B2b4A8e737BeA9b1'), Web3.to_checksum_address('0x1f9840a85d5aF5bf1D1762F925BDADdC4201F984'), 19881602, (1461501637330902918203684832716283019655932542975, 1718529375, 1)),  # noqa
        (Web3.to_checksum_address('0x1944922D86209F7e5F3c88F2B2b4A8e737BeA9b1'), Web3.to_checksum_address('0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48'), 19881602, (0, 0, 0)),  # noqa
    )
)
def test_fetch_permit2_allowance(wallet, token, block_identifier, expected_result, codec_rpc):
    amount, expiration, nonce = codec_rpc.fetch_permit2_allowance(
        wallet,
        token,
        spender=Web3.to_checksum_address("0x3fC91A3afd70395Cd496C647d5a6CC9D4B2b7FAD"),  # old UR address
        block_identifier=block_identifier
    )
    assert (amount, expiration, nonce) == expected_result



"""
Factory that builds the UR function ABIs used by the Uniswap Universal Router Codec

* Author: Elnaril (https://www.fiverr.com/elnaril, https://github.com/Elnaril).
* License: MIT.
* Doc: https://github.com/Elnaril/uniswap-universal-router-decoder
"""
from __future__ import annotations

from io import BytesIO
from typing import (
    Any,
    cast,
    Dict,
    List,
    Optional,
    Sequence,
    Union,
)

from eth_abi import encode
from eth_abi.registry import registry
from eth_utils import keccak
from web3 import Web3

from uniswap_universal_router_decoder._enums import (
    MiscFunctions,
    RouterFunction,
    V4Actions,
)


def _get_types_from_list(type_list: List[Any]) -> List[str]:
    types = []
    for item in type_list:
        if item["type"][:5] == "tuple":
            brackets = item["type"][5:]
            types.append(f"({','.join(_get_types_from_list(item['components']))}){brackets}")
        else:
            types.append(item["type"])
    return types


def build_abi_type_list(abi_dict: Dict[str, Any]) -> List[str]:
    return _get_types_from_list(abi_dict["inputs"])


class FunctionABI:
    def __init__(self, inputs: List[Any], name: str, _type: str) -> None:
        self.inputs = inputs
        self.name = name
        self.type = _type

    def get_abi(self) -> Dict[str, Any]:
        return {"inputs": self.inputs, "name": self.name, "type": self.type}

    def get_struct_abi(self) -> Dict[str, Any]:
        result = self.get_abi()
        result["components"] = result.pop("inputs")
        return result

    def get_full_abi(self) -> List[Dict[str, Any]]:
        return [self.get_abi()]

    def get_abi_types(self) -> List[str]:
        return build_abi_type_list(self.get_abi())

    def get_signature(self) -> str:
        return f"{self.name}({','.join(self.get_abi_types())})"

    def get_selector(self) -> bytes:
        return keccak(text=self.get_signature())[:4]

    def encode(self, args: Sequence[Any]) -> bytes:
        return encode(self.get_abi_types(), args)


ABIMap = Dict[Union[MiscFunctions, RouterFunction, V4Actions], FunctionABI]


class FunctionABIBuilder:
    def __init__(self, fct_name: str, _type: str = "function") -> None:
        self.abi = FunctionABI(inputs=[], name=fct_name, _type=_type)

    def add_address(self, arg_name: str) -> FunctionABIBuilder:
        self.abi.inputs.append({"name": arg_name, "type": "address"})
        return self

    def add_uint256(self, arg_name: str) -> FunctionABIBuilder:
        self.abi.inputs.append({"name": arg_name, "type": "uint256"})
        return self

    def add_uint160(self, arg_name: str) -> FunctionABIBuilder:
        self.abi.inputs.append({"name": arg_name, "type": "uint160"})
        return self

    def add_uint48(self, arg_name: str) -> FunctionABIBuilder:
        self.abi.inputs.append({"name": arg_name, "type": "uint48"})
        return self

    def add_uint24(self, arg_name: str) -> FunctionABIBuilder:
        self.abi.inputs.append({"name": arg_name, "type": "uint24"})
        return self

    def add_int24(self, arg_name: str) -> FunctionABIBuilder:
        self.abi.inputs.append({"name": arg_name, "type": "int24"})
        return self

    def add_uint128(self, arg_name: str) -> FunctionABIBuilder:
        self.abi.inputs.append({"name": arg_name, "type": "uint128"})
        return self

    def add_address_array(self, arg_name: str) -> FunctionABIBuilder:
        self.abi.inputs.append({"name": arg_name, "type": "address[]"})
        return self

    def add_bool(self, arg_name: str) -> FunctionABIBuilder:
        self.abi.inputs.append({"name": arg_name, "type": "bool"})
        return self

    def build(self) -> FunctionABI:
        return self.abi

    @staticmethod
    def create_struct_array(arg_name: str) -> FunctionABIBuilder:
        return FunctionABIBuilder(arg_name, "tuple[]")

    @staticmethod
    def create_struct(arg_name: str) -> FunctionABIBuilder:
        return FunctionABIBuilder(arg_name, "tuple")

    def add_struct(self, struct: FunctionABIBuilder) -> FunctionABIBuilder:
        self.abi.inputs.append(struct.abi.get_struct_abi())
        return self

    def add_struct_array(self, struct_array: FunctionABIBuilder) -> FunctionABIBuilder:
        self.abi.inputs.append(struct_array.abi.get_struct_abi())
        return self

    def add_bytes(self, arg_name: str) -> FunctionABIBuilder:
        self.abi.inputs.append({"name": arg_name, "type": "bytes"})
        return self

    def add_bytes_array(self, arg_name: str) -> FunctionABIBuilder:
        self.abi.inputs.append({"name": arg_name, "type": "bytes[]"})
        return self

    def add_v4_exact_input_params(self, arg_name: str = "params") -> FunctionABIBuilder:
        self.abi.inputs.append({"name": arg_name, "type": "ExactInputParams"})
        return self

    def add_v4_exact_output_params(self, arg_name: str = "params") -> FunctionABIBuilder:
        self.abi.inputs.append({"name": arg_name, "type": "ExactOutputParams"})
        return self


class _ABIBuilder:
    def __init__(self, w3: Optional[Web3] = None) -> None:
        self.w3 = w3 if w3 else Web3()
        self.abi_map = self.build_abi_map()
        if not registry.has_encoder("ExactInputParams"):
            registry.register("ExactInputParams", self.encode_v4_exact_input_params, self.decode_v4_exact_input_params)
        if not registry.has_encoder("ExactOutputParams"):
            registry.register(
                "ExactOutputParams",
                self.encode_v4_exact_output_params,
                self.decode_v4_exact_output_params,
            )

    def build_abi_map(self) -> ABIMap:
        abi_map: ABIMap = {
            # mapping between command identifier and function abi
            RouterFunction.V3_SWAP_EXACT_IN: self._build_v3_swap_exact_in(),
            RouterFunction.V3_SWAP_EXACT_OUT: self._build_v3_swap_exact_out(),
            RouterFunction.V2_SWAP_EXACT_IN: self._build_v2_swap_exact_in(),
            RouterFunction.V2_SWAP_EXACT_OUT: self._build_v2_swap_exact_out(),
            RouterFunction.PERMIT2_PERMIT: self._build_permit2_permit(),
            RouterFunction.WRAP_ETH: self._build_wrap_eth(),
            RouterFunction.UNWRAP_WETH: self._build_unwrap_weth(),
            RouterFunction.SWEEP: self._build_sweep(),
            RouterFunction.PAY_PORTION: self._build_pay_portion(),
            RouterFunction.TRANSFER: self._build_transfer(),
            RouterFunction.V4_SWAP: self._build_v4_swap(),
            RouterFunction.V4_INITIALIZE_POOL: self._build_v4_initialize_pool(),
            RouterFunction.V4_POSITION_MANAGER_CALL: self._build_modify_liquidities(),
            RouterFunction.PERMIT2_TRANSFER_FROM: self._build_permit2_transfer_from(),

            V4Actions.SWAP_EXACT_IN_SINGLE: self._build_v4_swap_exact_in_single(),
            V4Actions.MINT_POSITION: self._build_v4_mint_position(),
            V4Actions.SETTLE_PAIR: self._build_v4_settle_pair(),
            V4Actions.SETTLE: self._build_v4_settle(),
            V4Actions.CLOSE_CURRENCY: self._build_v4_close_currency(),
            V4Actions.SWEEP: self._build_v4_sweep(),
            V4Actions.TAKE_ALL: self._build_v4_take_all(),
            V4Actions.SETTLE_ALL: self._build_v4_settle_all(),
            V4Actions.SWAP_EXACT_IN: self._build_v4_swap_exact_in(),
            V4Actions.MINT_POSITION_FROM_DELTAS: self._build_v4_mint_position_from_deltas(),
            V4Actions.WRAP: self._build_v4_wrap_eth(),
            V4Actions.UNWRAP: self._build_v4_unwrap_weth(),
            V4Actions.SWAP_EXACT_OUT_SINGLE: self._build_v4_swap_exact_out_single(),
            V4Actions.SWAP_EXACT_OUT: self._build_v4_swap_exact_out(),
            V4Actions.TAKE_PAIR: self._build_v4_take_pair(),
            V4Actions.CLEAR_OR_TAKE: self._build_v4_clear_or_take(),
            V4Actions.TAKE_PORTION: self._build_v4_take_portion(),
            V4Actions.TAKE: self._build_v4_take(),

            MiscFunctions.EXECUTE: self._build_execute(),
            MiscFunctions.EXECUTE_WITH_DEADLINE: self._build_execute_with_deadline(),
            MiscFunctions.UNLOCK_DATA: self._build_unlock_data(),
            MiscFunctions.V4_POOL_ID: self._build_v4_pool_id(),
            MiscFunctions.STRICT_V4_SWAP_EXACT_IN: self._build_strict_v4_swap_exact_in(),
            MiscFunctions.STRICT_V4_SWAP_EXACT_OUT: self._build_strict_v4_swap_exact_out(),
        }
        return abi_map

    @staticmethod
    def _build_v2_swap_exact_in() -> FunctionABI:
        builder = FunctionABIBuilder(RouterFunction.V2_SWAP_EXACT_IN.name)
        builder.add_address("recipient").add_uint256("amountIn").add_uint256("amountOutMin").add_address_array("path")
        return builder.add_bool("payerIsSender").build()

    @staticmethod
    def _build_permit2_permit() -> FunctionABI:
        builder = FunctionABIBuilder(RouterFunction.PERMIT2_PERMIT.name)
        inner_struct = builder.create_struct("details")
        inner_struct.add_address("token").add_uint160("amount").add_uint48("expiration").add_uint48("nonce")
        outer_struct = builder.create_struct("struct")
        outer_struct.add_struct(inner_struct).add_address("spender").add_uint256("sigDeadline")
        return builder.add_struct(outer_struct).add_bytes("data").build()

    @staticmethod
    def _build_unwrap_weth() -> FunctionABI:
        builder = FunctionABIBuilder(RouterFunction.UNWRAP_WETH.name)
        return builder.add_address("recipient").add_uint256("amountMin").build()

    @staticmethod
    def _build_v3_swap_exact_in() -> FunctionABI:
        builder = FunctionABIBuilder(RouterFunction.V3_SWAP_EXACT_IN.name)
        builder.add_address("recipient").add_uint256("amountIn").add_uint256("amountOutMin").add_bytes("path")
        return builder.add_bool("payerIsSender").build()

    @staticmethod
    def _build_wrap_eth() -> FunctionABI:
        builder = FunctionABIBuilder(RouterFunction.WRAP_ETH.name)
        return builder.add_address("recipient").add_uint256("amountMin").build()

    @staticmethod
    def _build_v2_swap_exact_out() -> FunctionABI:
        builder = FunctionABIBuilder(RouterFunction.V2_SWAP_EXACT_OUT.name)
        builder.add_address("recipient").add_uint256("amountOut").add_uint256("amountInMax").add_address_array("path")
        return builder.add_bool("payerIsSender").build()

    @staticmethod
    def _build_v3_swap_exact_out() -> FunctionABI:
        builder = FunctionABIBuilder(RouterFunction.V3_SWAP_EXACT_OUT.name)
        builder.add_address("recipient").add_uint256("amountOut").add_uint256("amountInMax").add_bytes("path")
        return builder.add_bool("payerIsSender").build()

    @staticmethod
    def _build_sweep() -> FunctionABI:
        builder = FunctionABIBuilder(RouterFunction.SWEEP.name)
        return builder.add_address("token").add_address("recipient").add_uint256("amountMin").build()

    @staticmethod
    def _build_pay_portion() -> FunctionABI:
        builder = FunctionABIBuilder(RouterFunction.PAY_PORTION.name)
        return builder.add_address("token").add_address("recipient").add_uint256("bips").build()

    @staticmethod
    def _build_transfer() -> FunctionABI:
        builder = FunctionABIBuilder(RouterFunction.TRANSFER.name)
        return builder.add_address("token").add_address("recipient").add_uint256("value").build()

    @staticmethod
    def _build_v4_swap() -> FunctionABI:
        builder = FunctionABIBuilder(RouterFunction.V4_SWAP.name)
        return builder.add_bytes("actions").add_bytes_array("params").build()

    @staticmethod
    def _v4_pool_key_struct_builder() -> FunctionABIBuilder:
        builder = FunctionABIBuilder.create_struct("PoolKey")
        builder.add_address("currency0").add_address("currency1").add_uint24("fee").add_int24("tickSpacing")
        return builder.add_address("hooks")

    @staticmethod
    def _build_v4_swap_exact_in_single() -> FunctionABI:
        builder = FunctionABIBuilder(V4Actions.SWAP_EXACT_IN_SINGLE.name)
        pool_key = _ABIBuilder._v4_pool_key_struct_builder()
        outer_struct = builder.create_struct("exact_in_single_params")
        outer_struct.add_struct(pool_key).add_bool("zeroForOne").add_uint128("amountIn").add_uint128("amountOutMinimum")
        outer_struct.add_bytes("hookData")
        return builder.add_struct(outer_struct).build()

    @staticmethod
    def _build_v4_initialize_pool() -> FunctionABI:
        builder = FunctionABIBuilder(RouterFunction.V4_INITIALIZE_POOL.name)
        pool_key = _ABIBuilder._v4_pool_key_struct_builder()
        return builder.add_struct(pool_key).add_uint256("sqrtPriceX96").build()

    @staticmethod
    def _build_modify_liquidities() -> FunctionABI:
        builder = FunctionABIBuilder("modifyLiquidities")
        return builder.add_bytes("unlockData").add_uint256("deadline").build()

    @staticmethod
    def _build_unlock_data() -> FunctionABI:
        builder = FunctionABIBuilder("unlockData")
        return builder.add_bytes("actions").add_bytes_array("params").build()

    @staticmethod
    def _build_v4_mint_position() -> FunctionABI:
        builder = FunctionABIBuilder(V4Actions.MINT_POSITION.name)
        pool_key = _ABIBuilder._v4_pool_key_struct_builder()
        builder.add_struct(pool_key).add_int24("tickLower").add_int24("tickUpper").add_uint256("liquidity")
        builder.add_uint128("amount0Max").add_uint128("amount1Max").add_address("recipient").add_bytes("hookData")
        return builder.build()

    @staticmethod
    def _build_v4_settle_pair() -> FunctionABI:
        builder = FunctionABIBuilder(V4Actions.SETTLE_PAIR.name)
        return builder.add_address("currency0").add_address("currency1").build()

    @staticmethod
    def _build_v4_settle() -> FunctionABI:
        builder = FunctionABIBuilder(V4Actions.SETTLE.name)
        return builder.add_address("currency").add_uint256("amount").add_bool("payerIsUser").build()

    @staticmethod
    def _build_v4_close_currency() -> FunctionABI:
        builder = FunctionABIBuilder(V4Actions.CLOSE_CURRENCY.name)
        return builder.add_address("currency").build()

    @staticmethod
    def _build_v4_sweep() -> FunctionABI:
        builder = FunctionABIBuilder(V4Actions.SWEEP.name)
        return builder.add_address("currency").add_address("to").build()

    @staticmethod
    def _build_permit2_transfer_from() -> FunctionABI:
        builder = FunctionABIBuilder(RouterFunction.PERMIT2_TRANSFER_FROM.name)
        return builder.add_address("token").add_address("recipient").add_uint256("amount").build()

    @staticmethod
    def _build_v4_take_all() -> FunctionABI:
        builder = FunctionABIBuilder(V4Actions.TAKE_ALL.name)
        return builder.add_address("currency").add_uint256("minAmount").build()

    @staticmethod
    def _build_v4_settle_all() -> FunctionABI:
        builder = FunctionABIBuilder(V4Actions.SETTLE_ALL.name)
        return builder.add_address("currency").add_uint256("maxAmount").build()

    @staticmethod
    def _build_execute() -> FunctionABI:
        builder = FunctionABIBuilder("execute")
        return builder.add_bytes("commands").add_bytes_array("inputs").build()

    @staticmethod
    def _build_execute_with_deadline() -> FunctionABI:
        builder = FunctionABIBuilder("execute")
        return builder.add_bytes("commands").add_bytes_array("inputs").add_uint256("deadline").build()

    @staticmethod
    def _build_v4_pool_id() -> FunctionABI:
        builder = FunctionABIBuilder("v4_pool_id")
        pool_key = _ABIBuilder._v4_pool_key_struct_builder()
        return builder.add_struct(pool_key).build()

    @staticmethod
    def _v4_path_key_struct_array_builder() -> FunctionABIBuilder:
        builder = FunctionABIBuilder.create_struct_array("PathKeys")
        builder.add_address("intermediateCurrency").add_uint24("fee").add_int24("tickSpacing")
        return builder.add_address("hooks").add_bytes("hookData")

    @staticmethod
    def _build_strict_v4_swap_exact_in() -> FunctionABI:
        builder = FunctionABIBuilder(MiscFunctions.STRICT_V4_SWAP_EXACT_IN.name)
        builder.add_address("currencyIn")
        builder.add_struct_array(_ABIBuilder._v4_path_key_struct_array_builder())
        return builder.add_uint128("amountIn").add_uint128("amountOutMinimum").build()

    @staticmethod
    def _build_v4_mint_position_from_deltas() -> FunctionABI:
        builder = FunctionABIBuilder(V4Actions.MINT_POSITION_FROM_DELTAS.name)
        pool_key = _ABIBuilder._v4_pool_key_struct_builder()
        builder.add_struct(pool_key).add_int24("tickLower").add_int24("tickUpper")
        builder.add_uint128("amount0Max").add_uint128("amount1Max").add_address("recipient").add_bytes("hookData")
        return builder.build()

    @staticmethod
    def _build_v4_wrap_eth() -> FunctionABI:
        builder = FunctionABIBuilder(V4Actions.WRAP.name)
        return builder.add_uint256("amount").build()

    @staticmethod
    def _build_v4_unwrap_weth() -> FunctionABI:
        builder = FunctionABIBuilder(V4Actions.UNWRAP.name)
        return builder.add_uint256("amount").build()

    @staticmethod
    def _build_v4_swap_exact_out_single() -> FunctionABI:
        builder = FunctionABIBuilder(V4Actions.SWAP_EXACT_OUT_SINGLE.name)
        pool_key = _ABIBuilder._v4_pool_key_struct_builder()
        outer_struct = builder.create_struct("exact_out_single_params")
        outer_struct.add_struct(pool_key).add_bool("zeroForOne").add_uint128("amountOut").add_uint128("amountInMaximum")
        outer_struct.add_bytes("hookData")
        return builder.add_struct(outer_struct).build()

    @staticmethod
    def _build_strict_v4_swap_exact_out() -> FunctionABI:
        builder = FunctionABIBuilder(MiscFunctions.STRICT_V4_SWAP_EXACT_OUT.name)
        builder.add_address("currencyOut")
        builder.add_struct_array(_ABIBuilder._v4_path_key_struct_array_builder())
        return builder.add_uint128("amountOut").add_uint128("amountInMaximum").build()

    @staticmethod
    def _build_v4_take_pair() -> FunctionABI:
        builder = FunctionABIBuilder(V4Actions.TAKE_PAIR.name)
        return builder.add_address("currency0").add_address("currency1").add_address("recipient").build()

    @staticmethod
    def _build_v4_clear_or_take() -> FunctionABI:
        builder = FunctionABIBuilder(V4Actions.CLEAR_OR_TAKE.name)
        return builder.add_address("currency").add_uint256("amountMax").build()

    @staticmethod
    def _build_v4_take_portion() -> FunctionABI:
        builder = FunctionABIBuilder(V4Actions.TAKE_PORTION.name)
        return builder.add_address("currency").add_address("recipient").add_uint256("bips").build()

    @staticmethod
    def _build_v4_take() -> FunctionABI:
        builder = FunctionABIBuilder(V4Actions.TAKE.name)
        return builder.add_address("currency").add_address("recipient").add_uint256("amount").build()

    @staticmethod
    def _build_v4_swap_exact_in() -> FunctionABI:
        builder = FunctionABIBuilder(V4Actions.SWAP_EXACT_IN.name)
        return builder.add_v4_exact_input_params().build()

    def decode_v4_exact_input_params(self, stream: BytesIO) -> Dict[str, Any]:
        fct_abi = self.abi_map[MiscFunctions.STRICT_V4_SWAP_EXACT_IN]
        raw_data = stream.read()
        sub_contract = self.w3.eth.contract(abi=fct_abi.get_full_abi())
        fct_name, decoded_params = sub_contract.decode_function_input(fct_abi.get_selector() + raw_data[32:])
        return cast(Dict[str, Any], decoded_params)

    def encode_v4_exact_input_params(self, args: Sequence[Any]) -> bytes:
        fct_abi = self.abi_map[MiscFunctions.STRICT_V4_SWAP_EXACT_IN]
        encoded_data = 0x20.to_bytes(32, "big") + encode(fct_abi.get_abi_types(), args)
        return encoded_data

    @staticmethod
    def _build_v4_swap_exact_out() -> FunctionABI:
        builder = FunctionABIBuilder(V4Actions.SWAP_EXACT_OUT.name)
        return builder.add_v4_exact_output_params().build()

    def decode_v4_exact_output_params(self, stream: BytesIO) -> Dict[str, Any]:
        fct_abi = self.abi_map[MiscFunctions.STRICT_V4_SWAP_EXACT_OUT]
        raw_data = stream.read()
        sub_contract = self.w3.eth.contract(abi=fct_abi.get_full_abi())
        fct_name, decoded_params = sub_contract.decode_function_input(fct_abi.get_selector() + raw_data[32:])
        return cast(Dict[str, Any], decoded_params)

    def encode_v4_exact_output_params(self, args: Sequence[Any]) -> bytes:
        fct_abi = self.abi_map[MiscFunctions.STRICT_V4_SWAP_EXACT_OUT]
        encoded_data = 0x20.to_bytes(32, "big") + encode(fct_abi.get_abi_types(), args)
        return encoded_data



import pytest
from web3 import Web3
from web3.types import HexStr

from uniswap_universal_router_decoder import RouterCodec


# Test Decode Trx + Input
trx_hash_01 = HexStr("0x52e63b75f41a352ad9182f9e0f923c8557064c3b1047d1778c1ea5b11b979dd9")
expected_function_names_01 = ("PERMIT2_PERMIT", "V2_SWAP_EXACT_IN")

trx_hash_02 = HexStr("0x3247555a5dbc877ade17c4b49362bc981af5fb5064e0b3cbd91411e085fe3093")
expected_function_names_02 = ("V3_SWAP_EXACT_IN", "UNWRAP_WETH")

trx_hash_03 = HexStr("0x889b34a27b730dd664cd71579b4310522c3b495fb34f17f08d1131c0cec651fa")
expected_function_names_03 = ("WRAP_ETH", "V2_SWAP_EXACT_OUT", "UNWRAP_WETH")

trx_hash_04 = HexStr("0xf99ac4237df313794747759550db919b37d7c8a67d4a7e12be8f5cbaacd51376")
expected_function_names_04 = ("WRAP_ETH", "V2_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "UNWRAP_WETH")

trx_hash_05 = HexStr("0x47c0f1dd13edf9f1608f9f34bdba9ad40cb95dd081033cad69f5b88e451b4b55")
expected_function_names_05 = (None, "SWEEP")

trx_hash_06 = HexStr("0xe648089f71b2d2e7b70bdcbfdcfeecce6c5248b8eb64b2c79089b7c74c835a45")
expected_function_names_06 = ("V3_SWAP_EXACT_IN", "V3_SWAP_EXACT_IN", "V3_SWAP_EXACT_IN", "SWEEP")

trx_hash_07 = HexStr("0xb5a64e9935b46282080d9198f4478a4c4c1993d590eab8daa0f220c0dca5fe33")
expected_function_names_07 = ("V3_SWAP_EXACT_IN", "PAY_PORTION", "UNWRAP_WETH")

trx_hash_08 = HexStr("0x62176a906ef7f178814a0924d390082053bd8992c2902f436756194693644c21")
expected_function_names_08 = ("WRAP_ETH", "V2_SWAP_EXACT_OUT", "PAY_PORTION", "SWEEP", "UNWRAP_WETH")

trx_hash_09 = HexStr("0x2b6af8ef8fe18829a0fcf2b0f391c55daf76f53bb68369ecaefdb1f38045f919")
expected_function_names_09 = ("PERMIT2_PERMIT", "V2_SWAP_EXACT_IN", "V2_SWAP_EXACT_IN", "V3_SWAP_EXACT_IN", "V2_SWAP_EXACT_IN", "V3_SWAP_EXACT_IN", "SWEEP")  # noqa

trx_hash_10 = HexStr("0x586d51e2f92bd16573f0e7e302755ed02b7c2a4b721d63f46bcdcf7179d2f40e")
expected_function_names_10 = ("WRAP_ETH", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", "V3_SWAP_EXACT_OUT", None, "UNWRAP_WETH", "SWEEP")  # noqa


@pytest.mark.parametrize(
    "trx_hash, use_w3, expected_fct_names",
    (
        (trx_hash_01, False, expected_function_names_01),
        (trx_hash_02, True, expected_function_names_02),
        (trx_hash_03, True, expected_function_names_03),
        (trx_hash_04, True, expected_function_names_04),
        (trx_hash_05, True, expected_function_names_05),
        (trx_hash_06, True, expected_function_names_06),
        (trx_hash_07, True, expected_function_names_07),
        (trx_hash_08, True, expected_function_names_08),
        (trx_hash_09, True, expected_function_names_09),
        (trx_hash_10, True, expected_function_names_10),
    )
)
def test_decode_transaction(trx_hash, use_w3, expected_fct_names, w3, rpc_url):
    if use_w3:
        codec_w3 = RouterCodec(w3=w3)
    else:
        codec_w3 = RouterCodec(rpc_endpoint=rpc_url)

    decoded_trx = codec_w3.decode.transaction(trx_hash)
    command_inputs = decoded_trx["decoded_input"]["inputs"]
    assert len(command_inputs) == len(expected_fct_names)
    for i, expected_name in enumerate(expected_fct_names):
        if expected_name:
            assert expected_name == command_inputs[i][0].fn_name
            assert command_inputs[i][2]['revert_on_fail'] is True
        else:
            assert isinstance(command_inputs[i], str)
            int(command_inputs[i], 16)  # check the str is actually a hex


# Test Decode V3 Path

expected_parsed_path_02 = (
    Web3.to_checksum_address("0x4d224452801ACEd8B2F0aebE155379bb5D594381"),
    3000,
    Web3.to_checksum_address("0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"),
)

expected_parsed_path_04 = (
    Web3.to_checksum_address("0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"),
    3000,
    Web3.to_checksum_address("0xf3dcbc6D72a4E1892f7917b7C43b74131Df8480e"),
)

trx_hash_06 = HexStr("0x4a23b8ca6e15be1e61554d67bc5868b2fd8e91a97124b0fb31d39ce1a921bc62")
expected_parsed_path_06 = (
    Web3.to_checksum_address("0xdAC17F958D2ee523a2206206994597C13D831ec7"),
    500,
    Web3.to_checksum_address("0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"),
    3000,
    Web3.to_checksum_address("0xABe580E7ee158dA464b51ee1a83Ac0289622e6be"),
)

trx_hash_07 = HexStr("0xe61da48803242cb323056935737e8b430025ebe20af916031c2d53a1f1b2a844")
expected_parsed_path_07 = (
    Web3.to_checksum_address("0x0f51bb10119727a7e5eA3538074fb341F56B09Ad"),
    10000,
    Web3.to_checksum_address("0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48"),
    500,
    Web3.to_checksum_address("0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"),
)


@pytest.mark.parametrize(
    "trx_hash, fn_name, expected_parsed_path, expected_exception",
    (
        (trx_hash_02, "V3_SWAP_EXACT_IN", expected_parsed_path_02, None),
        (trx_hash_04, "V3_SWAP_EXACT_OUT", expected_parsed_path_04, None),
        (trx_hash_04, "V2_SWAP_EXACT_OUT", None, ValueError),
        (trx_hash_06, "V3_SWAP_EXACT_OUT", expected_parsed_path_06, None),
        (trx_hash_07, "V3_SWAP_EXACT_IN", expected_parsed_path_07, None),
    )
)
def test_decode_v3_path(trx_hash, fn_name, expected_parsed_path, expected_exception, codec_rpc):
    decoded_trx = codec_rpc.decode.transaction(trx_hash)
    command_inputs = decoded_trx["decoded_input"]["inputs"]
    for command_input in command_inputs:
        if command_input[0].fn_name == fn_name:
            path_bytes = command_input[1]["path"]
            if expected_exception:
                with pytest.raises(expected_exception):
                    _ = codec_rpc.decode.v3_path(fn_name, path_bytes)
            else:
                assert codec_rpc.decode.v3_path(fn_name, path_bytes) == expected_parsed_path
                assert codec_rpc.decode.v3_path(fn_name, path_bytes.hex()) == expected_parsed_path
            break
    else:
        raise ValueError(f"No fn_name {fn_name} found in the decoded command inputs for trx {trx_hash}")


contract_error_0 = "0x00000000"
contract_error_1 = "0x2c4029e9000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000400000000000000000000000000000000000000000000000000000000000000000"  # noqa
contract_error_2 = "0x5d1d0f9f"
contract_error_3 = "0x2c4029e9000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000400000000000000000000000000000000000000000000000000000000000000024bfb22adf000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"  # noqa
contract_sub_error_3 = b'\xbf\xb2*\xdf\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'.hex()  # noqa
contract_error_4 = "0xf801e5250000000000000000000000000000000000000000000000000000000000000000"


@pytest.mark.parametrize(
    "contract_error, expected",
    (
        (contract_error_0, ("Unknown error", {})),
        (contract_error_1, ("ExecutionFailed(uint256,bytes)", {'commandIndex': 0, 'message': b''})),
        (contract_error_2, ("OnlyMintAllowed()", {})),
        (contract_error_3, ("ExecutionFailed(uint256,bytes)", {'commandIndex': 0, 'message': b'\xbf\xb2*\xdf\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'})),  # noqa
        (contract_sub_error_3, ("DeadlinePassed(uint256)", {'deadline': 0})),
        (contract_error_4, ("InvalidAction(bytes4)", {'action': b'\x00\x00\x00\x00'})),
    )
)
def test_contract_error(contract_error, expected):
    codec = RouterCodec()
    decoded_error = codec.decode.contract_error(contract_error)
    assert decoded_error == expected



from uniswap_universal_router_decoder._encoder import (
    PathKey,
    PoolKey,
)
from uniswap_universal_router_decoder._enums import (
    FunctionRecipient,
    TransactionSpeed,
    V4Constants,
)
from uniswap_universal_router_decoder.router_codec import RouterCodec


__all__ = ["FunctionRecipient", "PathKey", "PoolKey", "RouterCodec", "TransactionSpeed", "V4Constants"]



"""
Enums used by the Uniswap Universal Router Codec

* Author: Elnaril (https://www.fiverr.com/elnaril, https://github.com/Elnaril).
* License: MIT.
* Doc: https://github.com/Elnaril/uniswap-universal-router-decoder
"""


from enum import (
    auto,
    Enum,
)

from web3 import Web3
from web3.types import Wei


class RouterFunction(Enum):
    # https://docs.uniswap.org/contracts/universal-router/technical-reference#command
    V3_SWAP_EXACT_IN = 0
    V3_SWAP_EXACT_OUT = 1
    PERMIT2_TRANSFER_FROM = 2
    SWEEP = 4
    TRANSFER = 5
    PAY_PORTION = 6
    V2_SWAP_EXACT_IN = 8
    V2_SWAP_EXACT_OUT = 9
    PERMIT2_PERMIT = 10
    WRAP_ETH = 11
    UNWRAP_WETH = 12
    V4_SWAP = 16
    V4_INITIALIZE_POOL = 19
    V4_POSITION_MANAGER_CALL = 20


class FunctionRecipient(Enum):
    """
    SENDER: When the function recipient is the sender

    ROUTER: When the function recipient is the router

    CUSTOM: When the function recipient is neither the trx sender nor the router
    """
    SENDER = "recipient is transaction sender"
    ROUTER = "recipient is universal router"
    CUSTOM = "recipient is custom"


class _RouterConstant(Enum):
    # https://github.com/Uniswap/universal-router/blob/main/contracts/libraries/Constants.sol
    MSG_SENDER = Web3.to_checksum_address("0x0000000000000000000000000000000000000001")
    ADDRESS_THIS = Web3.to_checksum_address("0x0000000000000000000000000000000000000002")
    ROUTER_BALANCE = Wei(2**255)
    FLAG_ALLOW_REVERT = 0x80
    COMMAND_TYPE_MASK = 0x3f


class TransactionSpeed(Enum):
    SLOW = 0
    AVERAGE = 1
    FAST = 2
    FASTER = 3


class V4Actions(Enum):
    # https://github.com/Uniswap/v4-periphery/blob/main/src/libraries/Actions.sol
    # Positions
    MINT_POSITION = 0x02
    MINT_POSITION_FROM_DELTAS = 0x05
    SETTLE_PAIR = 0x0d
    TAKE_PAIR = 0x11
    CLOSE_CURRENCY = 0x12
    CLEAR_OR_TAKE = 0x13
    SWEEP = 0x14
    WRAP = 0x15
    UNWRAP = 0x16

    # Swaps
    SWAP_EXACT_IN_SINGLE = 0x06
    SWAP_EXACT_IN = 0x07
    SWAP_EXACT_OUT_SINGLE = 0x08
    SWAP_EXACT_OUT = 0x09
    SETTLE_ALL = 0x0c
    TAKE_ALL = 0x0f
    TAKE_PORTION = 0x10

    # Common
    SETTLE = 0x0b
    TAKE = 0x0e


class V4Constants(Enum):
    OPEN_DELTA = 0
    CONTRACT_BALANCE = 0x8000000000000000000000000000000000000000000000000000000000000000


class MiscFunctions(Enum):
    EXECUTE = auto()
    EXECUTE_WITH_DEADLINE = auto()  # value = "execute" would be nice, but enum names and values must be unique
    UNLOCK_DATA = auto()
    V4_POOL_ID = auto()

    STRICT_V4_SWAP_EXACT_IN = auto()
    STRICT_V4_SWAP_EXACT_OUT = auto()



from statistics import quantiles
from typing import (
    cast,
    Sequence,
    Tuple,
)

from web3 import Web3
from web3.types import (
    BlockIdentifier,
    TxData,
    Wei,
)

from uniswap_universal_router_decoder._enums import TransactionSpeed


_speed_multiplier = {
    TransactionSpeed.SLOW: 1,
    TransactionSpeed.AVERAGE: 1,
    TransactionSpeed.FAST: 1.25,
    TransactionSpeed.FASTER: 1.5,
}


def compute_gas_fees(
        w3: Web3,
        trx_speed: TransactionSpeed = TransactionSpeed.FAST,
        block_identifier: BlockIdentifier = "latest") -> Tuple[Wei, Wei]:
    """
    Compute the priority_fee (maxPriorityFeePerGas) and max_fee_per_gas (maxFeePerGas) according to the given
    transaction 'speed'. All speeds will compute gas fees in order to try to place the transaction in the next block
    (without certainty, of course).
    Higher speeds would place the transaction higher in the block than lower ones.
    So, during strained conditions, the computed gas fees could be very high and should be double-checked before
    using them.

    :param w3: valid Web3 instance
    :param trx_speed: the desired transaction 'speed'
    :param block_identifier: the block number or identifier, default to 'latest'
    :return: the tuple (priority_fee, max_fee_per_gas)
    """
    block = w3.eth.get_block(block_identifier, True)
    transactions = cast(Sequence[TxData], block["transactions"])
    tips = [
        int(trx.get("maxPriorityFeePerGas", 0))
        for trx
        in transactions
        if trx.get("maxPriorityFeePerGas", 0) > 0
    ]
    if len(tips) < 3:
        priority_fee = 1
    else:
        quintiles = quantiles(tips, n=5, method="inclusive")
        priority_fee = int(quintiles[trx_speed.value] * _speed_multiplier[trx_speed])

    base_fee = block["baseFeePerGas"]
    max_fee_per_gas = int(base_fee * 1.5 + priority_fee)

    return Wei(priority_fee), Wei(max_fee_per_gas)


def compute_sqrt_price_x96(amount_0: Wei, amount_1: Wei) -> int:
    """
    Compute the sqrtPriceX96
    :param amount_0: amount of PoolKey.currency_0
    :param amount_1: amount of PoolKey.currency_1
    :return: floor(sqrt(amount_1 / amount_0) * 2^96)
    """
    return int(pow(amount_1 / amount_0, 1/2) * 2**96)


def convert_sqrt_price_x96(sqrt_price_x96: int) -> float:
    """
    Return the price
    :param sqrt_price_x96: the sqrtPriceX96
    :return: amount_1 / amount_0
    """
    return (sqrt_price_x96 / 2**96)**2



"""
Decoding part of the Uniswap Universal Router Codec

* Author: Elnaril (https://www.fiverr.com/elnaril, https://github.com/Elnaril).
* License: MIT.
* Doc: https://github.com/Elnaril/uniswap-universal-router-decoder
"""
from itertools import chain
import json
from typing import (
    Any,
    Dict,
    List,
    Sequence,
    Tuple,
    Union,
)

from eth_abi import decode
from eth_abi.exceptions import DecodingError
from web3 import Web3
from web3.contract.contract import BaseContractFunction
from web3.exceptions import Web3Exception
from web3.types import (
    ChecksumAddress,
    HexBytes,
    HexStr,
    TxData,
)

from uniswap_universal_router_decoder._abi_builder import (
    ABIMap,
    build_abi_type_list,
)
from uniswap_universal_router_decoder._constants import (
    _permit2_abi,
    _pool_manager_abi,
    _position_manager_abi,
    _router_abi,
)
from uniswap_universal_router_decoder._enums import (
    _RouterConstant,
    RouterFunction,
    V4Actions,
)


class _V4Decoder:
    def __init__(self, w3: Web3, abi_map: ABIMap) -> None:
        self._w3 = w3
        self._abi_map = abi_map
        self._pm_contract = w3.eth.contract(abi=_position_manager_abi)

    def _decode_v4_actions(
            self,
            actions: bytes,
            params: List[bytes]) -> List[Tuple[BaseContractFunction, Dict[str, Any]]]:
        if len(actions) != len(params):
            raise ValueError(f"Number of actions {len(actions)} is different from number of params: {len(params)}")

        decoded_params = []
        for i, action in enumerate(actions):
            try:
                abi_mapping = self._abi_map[V4Actions(action)]
                data = abi_mapping.get_selector() + params[i]
                sub_contract = self._w3.eth.contract(abi=abi_mapping.get_full_abi())
                decoded_params.append(sub_contract.decode_function_input(data))
            except (ValueError, KeyError, DecodingError):
                decoded_params.append(params[i].hex())
        return decoded_params

    def decode_v4_swap(self, actions: bytes, params: List[bytes]) -> List[Tuple[BaseContractFunction, Dict[str, Any]]]:
        return self._decode_v4_actions(actions, params)

    def decode_v4_pm_call(self, encoded_input: bytes) -> Dict[str, Any]:
        actions, params = decode(["bytes", "bytes[]"], encoded_input)
        return {"actions": actions, "params": self._decode_v4_actions(actions, params)}


class _Decoder:
    def __init__(self, w3: Web3, abi_map: ABIMap) -> None:
        self._w3 = w3
        self._router_contract = self._w3.eth.contract(abi=_router_abi)
        self._abi_map = abi_map
        self._v4_decoder = _V4Decoder(w3, abi_map)

    def function_input(self, input_data: Union[HexStr, HexBytes]) -> Tuple[BaseContractFunction, Dict[str, Any]]:
        """
        Decode the data sent to an UR function

        :param input_data: the transaction 'input' data
        :return: The decoded data if the function has been implemented.
        """
        fct_name, decoded_input = self._router_contract.decode_function_input(input_data)
        # returns (execute as basecontractfunction, {commands as bytes, inputs as seq of bytes, deadline as int})
        command = decoded_input["commands"]
        command_input = decoded_input["inputs"]
        decoded_command_input = []
        for i, b in enumerate(command):
            # iterating over bytes produces integers
            command_function = b & _RouterConstant.COMMAND_TYPE_MASK.value
            try:
                abi_mapping = self._abi_map[RouterFunction(command_function)]
                if b == RouterFunction.V4_POSITION_MANAGER_CALL.value:
                    data = command_input[i]
                else:
                    data = abi_mapping.get_selector() + command_input[i]
                sub_contract = self._w3.eth.contract(abi=abi_mapping.get_full_abi())
                revert_on_fail = not bool(b & _RouterConstant.FLAG_ALLOW_REVERT.value)
                decoded_fct_name, decoded_fct_params = sub_contract.decode_function_input(data)
                if b == RouterFunction.V4_SWAP.value:
                    decoded_command_input.append(
                        (
                            decoded_fct_name,
                            {
                                "actions": decoded_fct_params["actions"],
                                "params": self._v4_decoder.decode_v4_swap(
                                    decoded_fct_params["actions"],
                                    decoded_fct_params["params"],
                                ),
                            },
                            {"revert_on_fail": revert_on_fail},
                        )
                    )
                elif b == RouterFunction.V4_POSITION_MANAGER_CALL.value:
                    decoded_command_input.append(
                        (
                            decoded_fct_name,
                            {
                                "unlockData": self._v4_decoder.decode_v4_pm_call(decoded_fct_params["unlockData"]),
                                "deadline": decoded_fct_params["deadline"]
                            },
                            {"revert_on_fail": revert_on_fail},
                        )
                    )
                else:
                    decoded_command_input.append(
                        (
                            decoded_fct_name,
                            decoded_fct_params,
                            {"revert_on_fail": revert_on_fail}
                        )
                    )

            except (ValueError, KeyError, DecodingError):
                decoded_command_input.append(command_input[i].hex())
        decoded_input["inputs"] = decoded_command_input
        return fct_name, decoded_input

    def transaction(self, trx_hash: Union[HexBytes, HexStr]) -> Dict[str, Any]:
        """
        Get transaction details and decode the data used to call a UR function.

        ⚠ To use this method, the decoder must be built with a Web3 instance or a rpc endpoint address.

        :param trx_hash: the hash of the transaction sent to the UR
        :return: the transaction as a dict with the additional 'decoded_input' field
        """
        trx = self._get_transaction(trx_hash)
        fct_name, decoded_input = self.function_input(trx["input"])
        result_trx = dict(trx)
        result_trx["decoded_input"] = decoded_input
        return result_trx

    def _get_transaction(self, trx_hash: Union[HexBytes, HexStr]) -> TxData:
        return self._w3.eth.get_transaction(trx_hash)

    @staticmethod
    def v3_path(v3_fn_name: str, path: Union[bytes, str]) -> Tuple[Union[int, ChecksumAddress], ...]:
        """
        Decode a V3 router path

        :param v3_fn_name: V3_SWAP_EXACT_IN or V3_SWAP_EXACT_OUT only
        :param path: the V3 path as returned by decode_function_input() or decode_transaction()
        :return: a tuple of token addresses separated by the corresponding pool fees, first token being the 'in-token',
        last token being the 'out-token'
        """
        valid_fn_names = ("V3_SWAP_EXACT_IN", "V3_SWAP_EXACT_OUT")
        if v3_fn_name.upper() not in valid_fn_names:
            raise ValueError(f"v3_fn_name must be in {valid_fn_names}")
        path_str = path.hex() if isinstance(path, bytes) else str(path)
        path_str = path_str[2:] if path_str.startswith("0x") else path_str
        path_list: List[Union[int, ChecksumAddress]] = [Web3.to_checksum_address(path_str[0:40]), ]
        parsed_remaining_path: List[List[Union[int, ChecksumAddress]]] = [
            [
                int(path_str[40:][i:i + 6], 16),
                Web3.to_checksum_address(path_str[40:][i + 6:i + 46]),
            ]
            for i in range(0, len(path_str[40:]), 46)
        ]
        path_list.extend(list(chain.from_iterable(parsed_remaining_path)))

        if v3_fn_name.upper() == "V3_SWAP_EXACT_OUT":
            path_list.reverse()

        return tuple(path_list)

    def contract_error(
            self,
            contract_error: Union[str, HexStr],
            abis: Sequence[str] = (_permit2_abi, _pool_manager_abi, _position_manager_abi, _router_abi),
    ) -> Tuple[str, Dict[str, Any]]:
        """
        Decode contract custom errors.

        :param contract_error: the hexadecimal error, ex: '0x5d1d0f9f'
        :param abis: override the default abis which are permit2, v4 pool and position managers, and the UR
        :return: the decoded error if it's part of the abis, or 'Unknown error'
        """
        for abi in abis:
            try:
                json_abi = json.loads(abi)
                error_abi = []
                for item in json_abi:
                    if item["type"].lower() == "error":
                        item["type"] = "function"
                        error_abi.append(item)
                contract = self._w3.eth.contract(abi=error_abi)
                error, params = contract.decode_function_input(contract_error)
                return f"{error.fn_name}({','.join(build_abi_type_list(error.abi))})", params
            except (ValueError, Web3Exception):
                """The error is not defined in this ABI"""
        return "Unknown error", {}
